@@ -17,13 +17,12 @@
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <gtk/gtk.h>
-#include <gdk/gdkx.h>
 #include <X11/Xcursor/Xcursor.h>
 #include <X11/extensions/Xfixes.h>
 
-#include "mt-cursor.h"
 #include "mt-cursor-manager.h"
+#include "mt-cursor.h"
+#include "mt-common.h"
 
 static int xfixes_event = 0;
 static int xfixes_error = 0;
@@ -31,36 +30,33 @@ static int xfixes_error = 0;
 struct _MtCursorManagerPrivate
 {
     GHashTable *cache;
+    MtCursor   *current;
 };
 
 enum
 {
     CURSOR_CHANGED,
-    CACHE_CLEARED,
     LAST_SIGNAL
 };
 
 static guint signals[LAST_SIGNAL] = { 0, };
 
-static void            mt_cursor_manager_clear_cache  (GObject         *settings,
-                                                       GParamSpec      *pspec,
-                                                       MtCursorManager *manager);
-static GdkFilterReturn mt_cursor_manager_event_filter (GdkXEvent       *gdk_xevent,
-                                                       GdkEvent        *gdk_event,
-                                                       gpointer         data);
+static void            mt_cursor_manager_clear_cache        (GObject         *settings,
+                                                             GParamSpec      *pspec,
+                                                             MtCursorManager *manager);
+static GdkFilterReturn mt_cursor_manager_event_filter       (GdkXEvent       *gdk_xevent,
+                                                             GdkEvent        *gdk_event,
+                                                             MtCursorManager *manager);
+static void            mt_cursor_manager_add_current_cursor (MtCursorManager *manager,
+                                                             Display         *display);
 
 G_DEFINE_TYPE (MtCursorManager, mt_cursor_manager, G_TYPE_OBJECT)
 
-static Display *
-mt_cursor_manager_display (void)
-{
-    return GDK_DISPLAY ();
-}
-
 static void
 mt_cursor_manager_init (MtCursorManager *manager)
 {
     GtkSettings *gs;
+    Display *dpy;
 
     manager->priv = G_TYPE_INSTANCE_GET_PRIVATE (manager,
                                                  MT_TYPE_CURSOR_MANAGER,
@@ -71,16 +67,16 @@ mt_cursor_manager_init (MtCursorManager *manager)
                                                   g_free,
                                                   g_object_unref);
 
+    dpy = mt_common_get_xdisplay ();
+
     /* init XFixes extension */
-    if (XFixesQueryExtension (mt_cursor_manager_display (),
-                              &xfixes_event, &xfixes_error))
+    if (XFixesQueryExtension (dpy, &xfixes_event, &xfixes_error))
     {
-        XFixesSelectCursorInput (mt_cursor_manager_display (),
-                                 GDK_ROOT_WINDOW (),
+        XFixesSelectCursorInput (dpy, GDK_ROOT_WINDOW (),
                                  XFixesDisplayCursorNotifyMask);
     }
 
-    gdk_window_add_filter (NULL, mt_cursor_manager_event_filter, manager);
+    gdk_window_add_filter (NULL, (GdkFilterFunc) mt_cursor_manager_event_filter, manager);
 
     /* listen for cursor theme changes */
     gs = gtk_settings_get_default ();
@@ -88,6 +84,8 @@ mt_cursor_manager_init (MtCursorManager *manager)
                       G_CALLBACK (mt_cursor_manager_clear_cache), manager);
     g_signal_connect (gs, "notify::gtk-cursor-theme-size",
                       G_CALLBACK (mt_cursor_manager_clear_cache), manager);
+
+    mt_cursor_manager_add_current_cursor (manager, dpy);
 }
 
 static void
@@ -95,7 +93,7 @@ mt_cursor_manager_finalize (GObject *object)
 {
     MtCursorManagerPrivate *priv = MT_CURSOR_MANAGER (object)->priv;
 
-    gdk_window_remove_filter (NULL, mt_cursor_manager_event_filter, object);
+    gdk_window_remove_filter (NULL, (GdkFilterFunc) mt_cursor_manager_event_filter, object);
     g_hash_table_destroy (priv->cache);
 
     G_OBJECT_CLASS (mt_cursor_manager_parent_class)->finalize (object);
@@ -116,14 +114,6 @@ mt_cursor_manager_class_init (MtCursorManagerClass *klass)
                       g_cclosure_marshal_VOID__STRING,
                       G_TYPE_NONE, 1, G_TYPE_STRING);
 
-    signals[CACHE_CLEARED] =
-        g_signal_new (g_intern_static_string ("cache_cleared"),
-                      G_OBJECT_CLASS_TYPE (klass),
-                      G_SIGNAL_RUN_LAST,
-                      0, NULL, NULL,
-                      g_cclosure_marshal_VOID__VOID,
-                      G_TYPE_NONE, 0);
-
     g_type_class_add_private (klass, sizeof (MtCursorManagerPrivate));
 }
 
@@ -137,7 +127,7 @@ mt_cursor_manager_set_xcursor (MtCursor *cursor)
     gushort width, height;
     gushort xhot, yhot;
 
-    dpy = mt_cursor_manager_display ();
+    dpy = mt_common_get_xdisplay ();
     mt_cursor_get_dimension (cursor, &width, &height);
     ximage = XcursorImageCreate (width, height);
     xcursor = 0;
@@ -175,7 +165,7 @@ mt_cursor_manager_restore_cursor (gpointer key,
     mt_cursor_manager_set_xcursor (value);
 }
 
-static void
+static MtCursor *
 mt_cursor_manager_add_cursor (MtCursorManager   *manager,
                               XFixesCursorImage *image)
 {
@@ -197,7 +187,7 @@ mt_cursor_manager_add_cursor (MtCursorManager   *manager,
     }
     else
     {
-        pixels = (guchar *) image->pixels;
+        pixels = g_memdup (image->pixels, image->width * image->height * 4);
     }
 
     cursor = mt_cursor_new (image->name, pixels,
@@ -209,12 +199,28 @@ mt_cursor_manager_add_cursor (MtCursorManager   *manager,
                              g_strdup (image->name),
                              cursor);
     }
+
+    return cursor;
+}
+
+static void
+mt_cursor_manager_add_current_cursor (MtCursorManager *manager,
+                                      Display         *display)
+{
+    XFixesCursorImage *image;
+
+    image = XFixesGetCursorImage (display);
+    if (image)
+    {
+        manager->priv->current = mt_cursor_manager_add_cursor (manager, image);
+        XFree (image);
+    }
 }
 
 static GdkFilterReturn
-mt_cursor_manager_event_filter (GdkXEvent *gdk_xevent,
-                                GdkEvent  *gdk_event,
-                                gpointer   data)
+mt_cursor_manager_event_filter (GdkXEvent       *gdk_xevent,
+                                GdkEvent        *gdk_event,
+                                MtCursorManager *manager)
 {
     XEvent *xev = gdk_xevent;
 
@@ -224,15 +230,23 @@ mt_cursor_manager_event_filter (GdkXEvent *gdk_xevent,
 
         if (cn->cursor_name != None)
         {
-            MtCursorManager *manager = data;
-            XFixesCursorImage *image;
+            MtCursorManagerPrivate *priv = manager->priv;
+            char *name;
 
-            image = XFixesGetCursorImage (cn->display);
-            if (!mt_cursor_manager_lookup_cursor (manager, image->name))
-                mt_cursor_manager_add_cursor (manager, image);
+            name = XGetAtomName (cn->display, cn->cursor_name);
 
-            g_signal_emit (manager, signals[CURSOR_CHANGED], 0, image->name);
-            XFree (image);
+            if (!priv->current || !g_str_equal (name, mt_cursor_get_name (priv->current)))
+            {
+                priv->current = mt_cursor_manager_lookup_cursor (manager, name);
+
+                if (!priv->current)
+                {
+                    mt_cursor_manager_add_current_cursor (manager, cn->display);
+                }
+
+                g_signal_emit (manager, signals[CURSOR_CHANGED], 0, name);
+            }
+            XFree (name);
         }
     }
     return GDK_FILTER_CONTINUE;
@@ -244,7 +258,7 @@ mt_cursor_manager_clear_cache (GObject         *settings,
                                MtCursorManager *manager)
 {
     g_hash_table_remove_all (manager->priv->cache);
-    g_signal_emit (manager, signals[CACHE_CLEARED], 0);
+    manager->priv->current = NULL;
 }
 
 MtCursorManager *
@@ -253,34 +267,19 @@ mt_cursor_manager_get_default (void)
     static MtCursorManager *manager = NULL;
 
     if (!manager)
+    {
         manager = g_object_new (MT_TYPE_CURSOR_MANAGER, NULL);
-
+        g_object_add_weak_pointer (G_OBJECT (manager), (gpointer *) &manager);
+    }
     return manager;
 }
 
 MtCursor *
-mt_cursor_manager_current_cursor (MtCursorManager *manager)
+mt_cursor_manager_get_current_cursor (MtCursorManager *manager)
 {
-    XFixesCursorImage *image;
-    MtCursor *cursor;
-
     g_return_val_if_fail (MT_IS_CURSOR_MANAGER (manager), NULL);
 
-    image = XFixesGetCursorImage (mt_cursor_manager_display ());
-    cursor = NULL;
-
-    if (image->name && image->name[0] != '\0')
-    {
-        cursor = mt_cursor_manager_lookup_cursor (manager, image->name);
-        if (!cursor)
-        {
-            mt_cursor_manager_add_cursor (manager, image);
-            cursor = mt_cursor_manager_lookup_cursor (manager, image->name);
-        }
-    }
-    XFree (image);
-
-    return cursor;
+    return g_object_ref (manager->priv->current);
 }
 
 MtCursor *
